@@ -64,11 +64,12 @@ public interface ClientProtoUtils {
 
   static RaftRpcRequestProto.Builder toRaftRpcRequestProtoBuilder(RaftGroupMemberId requestorId, RaftPeerId replyId) {
     return toRaftRpcRequestProtoBuilder(requestorId.getPeerId().toByteString(),
-        replyId.toByteString(), requestorId.getGroupId(), null, null, null, 0);
+        replyId.toByteString(), requestorId.getGroupId(), null, false, null, null, 0);
   }
 
+  @SuppressWarnings("parameternumber")
   static RaftRpcRequestProto.Builder toRaftRpcRequestProtoBuilder(
-      ByteString requesterId, ByteString replyId, RaftGroupId groupId, Long callId,
+      ByteString requesterId, ByteString replyId, RaftGroupId groupId, Long callId, boolean toLeader,
       SlidingWindowEntry slidingWindowEntry, RoutingTable routingTable, long timeoutMs) {
     if (slidingWindowEntry == null) {
       slidingWindowEntry = SlidingWindowEntry.getDefaultInstance();
@@ -79,6 +80,7 @@ public interface ClientProtoUtils {
         .setReplyId(replyId)
         .setRaftGroupId(ProtoUtils.toRaftGroupIdProtoBuilder(groupId))
         .setCallId(Optional.ofNullable(callId).orElseGet(CallId::getDefault))
+        .setToLeader(toLeader)
         .setSlidingWindowEntry(slidingWindowEntry)
         .setTimeoutMs(timeoutMs);
 
@@ -89,11 +91,12 @@ public interface ClientProtoUtils {
     return b;
   }
 
+  @SuppressWarnings("parameternumber")
   static RaftRpcRequestProto.Builder toRaftRpcRequestProtoBuilder(
-      ClientId requesterId, RaftPeerId replyId, RaftGroupId groupId, long callId,
+      ClientId requesterId, RaftPeerId replyId, RaftGroupId groupId, long callId, boolean toLeader,
       SlidingWindowEntry slidingWindowEntry, RoutingTable routingTable, long timeoutMs) {
     return toRaftRpcRequestProtoBuilder(
-        requesterId.toByteString(), replyId.toByteString(), groupId, callId, slidingWindowEntry, routingTable,
+        requesterId.toByteString(), replyId.toByteString(), groupId, callId, toLeader, slidingWindowEntry, routingTable,
         timeoutMs);
   }
 
@@ -104,6 +107,7 @@ public interface ClientProtoUtils {
         request.getServerId(),
         request.getRaftGroupId(),
         request.getCallId(),
+        request.isToLeader(),
         request.getSlidingWindowEntry(),
         request.getRoutingTable(),
         request.getTimeoutMs());
@@ -151,9 +155,15 @@ public interface ClientProtoUtils {
     final RaftClientRequest.Type type = toRaftClientRequestType(p);
     final RaftRpcRequestProto request = p.getRpcRequest();
 
-    return RaftClientRequest.newBuilder()
-        .setClientId(ClientId.valueOf(request.getRequestorId()))
-        .setServerId(RaftPeerId.valueOf(request.getReplyId()))
+    final RaftClientRequest.Builder b = RaftClientRequest.newBuilder();
+
+    final RaftPeerId perrId = RaftPeerId.valueOf(request.getReplyId());
+    if (request.getToLeader()) {
+      b.setLeaderId(perrId);
+    } else {
+      b.setServerId(perrId);
+    }
+    return b.setClientId(ClientId.valueOf(request.getRequestorId()))
         .setGroupId(ProtoUtils.toRaftGroupId(request.getRaftGroupId()))
         .setCallId(request.getCallId())
         .setMessage(toMessage(p.getMessage()))
@@ -211,7 +221,7 @@ public interface ClientProtoUtils {
       long seqNum, ByteString content) {
     return RaftClientRequestProto.newBuilder()
         .setRpcRequest(toRaftRpcRequestProtoBuilder(
-            clientId, serverId, groupId, callId, ProtoUtils.toSlidingWindowEntry(seqNum, false), null, 0))
+            clientId, serverId, groupId, callId, false, ProtoUtils.toSlidingWindowEntry(seqNum, false), null, 0))
         .setWrite(WriteRequestTypeProto.getDefaultInstance())
         .setMessage(toClientMessageEntryProtoBuilder(content))
         .build();
