@@ -114,11 +114,23 @@ public interface RaftServerConfigKeys {
     SizeInBytes BYTE_LIMIT_DEFAULT = SizeInBytes.valueOf("64MB");
     static SizeInBytes byteLimit(RaftProperties properties) {
       return getSizeInBytes(properties::getSizeInBytes,
-          BYTE_LIMIT_KEY, BYTE_LIMIT_DEFAULT, getDefaultLog());
+          BYTE_LIMIT_KEY, BYTE_LIMIT_DEFAULT, getDefaultLog(), requireMinSizeInByte(SizeInBytes.ONE_MB));
     }
     static void setByteLimit(RaftProperties properties, SizeInBytes byteLimit) {
       setSizeInBytes(properties::set, BYTE_LIMIT_KEY, byteLimit, requireMin(1L));
     }
+
+    String FOLLOWER_GAP_RATIO_MAX_KEY = PREFIX + ".follower.gap.ratio.max";
+    // The valid range is [1, 0) and -1, -1 means disable this feature
+    double FOLLOWER_GAP_RATIO_MAX_DEFAULT = -1d;
+
+    static double followerGapRatioMax(RaftProperties properties) {
+      return getDouble(properties::getDouble, FOLLOWER_GAP_RATIO_MAX_KEY,
+          FOLLOWER_GAP_RATIO_MAX_DEFAULT, getDefaultLog(), requireMax(1d));
+    }
+    static void setFollowerGapRatioMax(RaftProperties properties, float ratio) {
+      setDouble(properties::setDouble, FOLLOWER_GAP_RATIO_MAX_KEY, ratio, requireMax(1d));
+    }
   }
 
   interface Watch {
@@ -183,9 +195,14 @@ public interface RaftServerConfigKeys {
       return getSizeInBytes(properties::getSizeInBytes,
           QUEUE_BYTE_LIMIT_KEY, QUEUE_BYTE_LIMIT_DEFAULT, getDefaultLog());
     }
+
+    @Deprecated
     static void setQueueByteLimit(RaftProperties properties, int queueSize) {
       setInt(properties::setInt, QUEUE_BYTE_LIMIT_KEY, queueSize, requireMin(1));
     }
+    static void setQueueByteLimit(RaftProperties properties, SizeInBytes byteLimit) {
+      setSizeInBytes(properties::set, QUEUE_BYTE_LIMIT_KEY, byteLimit, requireMin(1L));
+    }
 
     String PURGE_GAP_KEY = PREFIX + ".purge.gap";
     int PURGE_GAP_DEFAULT = 1024;
